import { useContext, useRef, useState, useEffect } from 'react';
import { ApolloContext } from 'react-apollo';
import { parser } from 'react-apollo/parser';
import isEqual from 'lodash/isEqual';

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(source, true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(source).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function observableQueryFields(observable) {
  if (!observable) return {};
  var fields = {
    variables: observable.variables,
    refetch: observable.refetch.bind(observable),
    fetchMore: observable.fetchMore.bind(observable),
    updateQuery: observable.updateQuery.bind(observable),
    startPolling: observable.startPolling.bind(observable),
    stopPolling: observable.stopPolling.bind(observable),
    subscribeToMore: observable.subscribeToMore.bind(observable)
  }; // TODO: Need to cast this because we improved the type of `updateQuery` to be parametric
  // on variables, while the type in Apollo client just has object.
  // Consider removing this when that is properly typed

  return fields;
}

function useQuery(query) {
  var props = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var {
    client,
    operations
  } = useContext(ApolloContext);
  var Observable = useRef(!props.skip ? client.watchQuery(_objectSpread2({
    query
  }, props)) : undefined);
  var Subscription = useRef();
  var [queryResult, _setQueryResult] = useState(_objectSpread2({
    data: undefined,
    error: undefined,
    loading: false,
    networkStatus: undefined
  }, observableQueryFields(Observable.current), {}, Observable.current && Observable.current.getCurrentResult()));

  function setQueryResult(newState) {
    _setQueryResult(_objectSpread2({}, observableQueryFields(Observable.current), {}, queryResult, {}, newState));
  }

  useEffect(() => {
    if (!props.skip) {
      if (Observable.current) {
        if (Subscription.current) Subscription.current.unsubscribe();
        Subscription.current = Observable.current && Observable.current.subscribe({
          next() {
            var prevResult = queryResult;
            var nextResult = Observable.current.getCurrentResult();
            if (!isEqual(nextResult.data, prevResult.data)) setQueryResult(nextResult);
          },

          error(err) {
            setQueryResult({
              error: err
            });
          }

        });
      } else {
        Observable.current = client.watchQuery(_objectSpread2({
          query
        }, props));
        setQueryResult({
          networkStatus: 1,
          loading: true,
          data: undefined,
          error: undefined
        });
      }
    }

    return () => {
      if (Subscription.current) Subscription.current.unsubscribe();
    };
  }, [props, query]);
  useEffect(() => {
    if (!Observable.current) return;
    Observable.current.setOptions(props);
  }, [props]);
  useEffect(() => {
    var operation = parser(query);
    operations.set(operation.name, {
      query,
      variables: props.variables
    });
  }, [query, props.variables]);
  return queryResult;
}

function useMutation(mutation, props) {
  var {
    client,
    operations
  } = useContext(ApolloContext);
  var [mutationResult, setMutationResult] = useState({
    loading: false,
    error: undefined,
    data: undefined
  });
  var [mutatorProps, setMutatorProps] = useState(props);
  var [runner, setRunner] = useState({
    counter: 0
  });
  useEffect(() => {
    var didCancel = false;

    function mutate() {
      return _mutate.apply(this, arguments);
    }

    function _mutate() {
      _mutate = _asyncToGenerator(function* () {
        try {
          if (!didCancel) setMutationResult(_objectSpread2({}, mutationResult, {
            error: undefined,
            loading: true
          }));
          var {
            refetchQueries
          } = mutatorProps;
          var parsedRefetchQueries = refetchQueries;

          if (refetchQueries && refetchQueries.length && Array.isArray(refetchQueries)) {
            parsedRefetchQueries = refetchQueries.map(x => {
              if (typeof x === 'string' && operations) return operations.get(x) || x;
              return x;
            });
          }

          var res = yield client.mutate(_objectSpread2({
            mutation
          }, mutatorProps, {
            refetchQueries: parsedRefetchQueries
          }));

          if (!didCancel) {
            setMutationResult(_objectSpread2({}, mutationResult, {
              loading: false,
              error: undefined,
              data: res.data
            }));
            runner.resolve(res);
          }
        } catch (err) {
          if (!didCancel) {
            setMutationResult(_objectSpread2({}, mutationResult, {
              loading: false,
              error: err,
              data: undefined
            }));
            runner.reject(err);
          }
        }
      });
      return _mutate.apply(this, arguments);
    }

    if (runner.counter !== 0) mutate();
    return () => {
      didCancel = true;
      if (runner.reject) runner.reject({
        canceled: true
      });
    };
  }, [runner.counter, mutatorProps, mutation]);

  function runMutation(runMutationProps) {
    return new Promise((resolve, reject) => {
      setMutatorProps(_objectSpread2({}, props, {}, runMutationProps));
      setRunner({
        counter: runner.counter + 1,
        resolve,
        reject
      });
    });
  }

  return [runMutation, mutationResult];
}

export { useMutation, useQuery };
